# Copyright © 2017 Tom Hacohen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from rest_framework import permissions
from journal.models import Journal, JournalMember


class IsOwnerOrReadOnly(permissions.BasePermission):
    """
    Custom permission to only allow owners of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        if request.method in permissions.SAFE_METHODS:
            return True

        return obj.owner == request.user


class IsJournalOwner(permissions.BasePermission):
    """
    Custom permission to only allow owners of a journal to view it
    """

    def has_permission(self, request, view):
        journal_uid = view.kwargs['journal_uid']
        try:
            journal = view.get_journal_queryset().get(uid=journal_uid)
            return journal.owner == request.user
        except Journal.DoesNotExist:
            # If the journal does not exist, we want to 404 later, not permission denied.
            return True


class IsMemberReadOnly(permissions.BasePermission):
    """
    Custom permission to make a journal read only if a read only member
    """

    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return True

        journal_uid = view.kwargs['journal_uid']
        try:
            journal = view.get_journal_queryset().get(uid=journal_uid)
            member = journal.members.get(user=request.user)
            return not member.readOnly
        except Journal.DoesNotExist:
            # If the journal does not exist, we want to 404 later, not permission denied.
            return True
        except JournalMember.DoesNotExist:
            # Not being a member means we are the owner.
            return True



# Copyright © 2017 Tom Hacohen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import base64

from django.db.models.expressions import RawSQL
from django.contrib.auth import get_user_model
from rest_framework import serializers
from . import models

User = get_user_model()


class BinaryBase64Field(serializers.Field):
    def to_representation(self, value):
        return base64.b64encode(value).decode('ascii')

    def to_internal_value(self, data):
        return base64.b64decode(data)


class JournalSerializer(serializers.ModelSerializer):
    content = BinaryBase64Field()
    owner = serializers.SlugRelatedField(
        slug_field=User.USERNAME_FIELD,
        read_only=True
    )
    key = serializers.SerializerMethodField('get_key_from_context')
    readOnly = serializers.SerializerMethodField('get_read_only_from_context')
    lastUid = serializers.SerializerMethodField('get_last_uid')

    class Meta:
        model = models.Journal
        fields = ('version', 'uid', 'content', 'owner', 'key', 'readOnly', 'lastUid')

    def get_key_from_context(self, obj):
        request = self.context.get('request', None)
        if request is not None:
            try:
                member = obj.members.get(user=request.user)
                serialized_member = JournalMemberSerializer(member)
                return serialized_member.data['key']
            except models.JournalMember.DoesNotExist:
                pass
        return None

    def get_read_only_from_context(self, obj):
        request = self.context.get('request', None)
        if request is not None:
            try:
                member = obj.members.get(user=request.user)
                return member.readOnly
            except models.JournalMember.DoesNotExist:
                pass
        return False

    def get_last_uid(self, obj):
        last = models.Entry.objects.filter(
                id=RawSQL('SELECT MAX(journal_entry.id) FROM journal_entry WHERE journal_entry.journal_id = %s GROUP BY journal_entry.journal_id', (obj.id, ))
            ).first()
        if last:
            return last.uid
        return None


class JournalUpdateSerializer(JournalSerializer):
    class Meta(JournalSerializer.Meta):
        fields = ('content', )


class EntrySerializer(serializers.ModelSerializer):
    content = BinaryBase64Field()

    class Meta:
        model = models.Entry
        fields = ('uid', 'content')


class UserInfoSerializer(serializers.ModelSerializer):
    content = BinaryBase64Field()
    pubkey = BinaryBase64Field()

    class Meta:
        model = models.UserInfo
        fields = ('version', 'pubkey', 'content')


class UserInfoPublicSerializer(UserInfoSerializer):
    class Meta(JournalSerializer.Meta):
        fields = ('version', 'pubkey')


class JournalMemberSerializer(serializers.ModelSerializer):
    user = serializers.SlugRelatedField(
        slug_field=User.USERNAME_FIELD,
        queryset=User.objects
    )
    key = BinaryBase64Field()

    class Meta:
        model = models.JournalMember
        fields = ('user', 'key', 'readOnly')



# Generated by Django 2.1.7 on 2019-02-13 11:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('journal', '0009_auto_20190211_1515'),
    ]

    operations = [
        migrations.AddField(
            model_name='journalmember',
            name='readOnly',
            field=models.BooleanField(default=False),
        ),
    ]



# Copyright © 2017 Tom Hacohen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

DEBUG = False

SECRET_KEY = 'fake-key'

ROOT_URLCONF = 'tests.test_urls'

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.sites',
    'django.contrib.staticfiles',
    'rest_framework',
    'journal.apps.JournalConfig',
    'tests',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates')
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'core.context_processors.extra',
            ],
        },
    },
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

REST_FRAMEWORK = {
    'TEST_REQUEST_DEFAULT_FORMAT': 'json',
}



# Generated by Django 2.1.6 on 2019-02-11 15:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('journal', '0008_remove_userinfo_deleted'),
    ]

    operations = [
        migrations.AlterField(
            model_name='entry',
            name='content',
            field=models.BinaryField(editable=True),
        ),
        migrations.AlterField(
            model_name='journal',
            name='content',
            field=models.BinaryField(editable=True),
        ),
        migrations.AlterField(
            model_name='journalmember',
            name='key',
            field=models.BinaryField(editable=True),
        ),
        migrations.AlterField(
            model_name='userinfo',
            name='content',
            field=models.BinaryField(editable=True),
        ),
        migrations.AlterField(
            model_name='userinfo',
            name='pubkey',
            field=models.BinaryField(editable=True),
        ),
    ]



# Copyright © 2017 Tom Hacohen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import sys  # noqa


class AppSettings:
    def __init__(self, prefix):
        self.prefix = prefix

    def import_from_str(self, name):
        from importlib import import_module

        p, m = name.rsplit('.', 1)

        mod = import_module(p)
        return getattr(mod, m)

    def _setting(self, name, dflt):
        from django.conf import settings
        return getattr(settings, self.prefix + name, dflt)

    @property
    def API_PERMISSIONS(self):
        perms = self._setting("API_PERMISSIONS", ('rest_framework.permissions.IsAuthenticated', ))
        ret = []
        for perm in perms:
            ret.append(self.import_from_str(perm))
        return ret

    @property
    def API_AUTHENTICATORS(self):
        perms = self._setting("API_AUTHENTICATORS", ('rest_framework.authentication.TokenAuthentication',
                                                     'rest_framework.authentication.SessionAuthentication'))
        ret = []
        for perm in perms:
            ret.append(self.import_from_str(perm))
        return ret


# Ugly? Guido recommends this himself ...
# http://mail.python.org/pipermail/python-ideas/2012-May/014969.html
app_settings = AppSettings('JOURNAL_')
app_settings.__name__ = __name__  # pylint: disable=attribute-defined-outside-init
sys.modules[__name__] = app_settings



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-17 16:16
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('journal', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='entry',
            options={'ordering': ['id']},
        ),
    ]
